;============================================================================
; (c) Copyright Elect Software International Inc., 1992, Toronto. Anyone can
; use this code for anything as long as it is not resold as a software
; development resource, as long as the copyright notice isn't removed, as
; long as changes are clearly marked as to authorship, and as long as users
; indemnify Elect from any liability.
; Comments welcome. Henrik Bechmann, CIS:72701,3717; Tel:416-534-8176.
;============================================================================

;============================================================================
;                            EVENT DESCRIPTION
;============================================================================
;
; EventMan (Event Manager) is a generic event manager for use in Paradox 4.0
; applications. The general organization of the scheme is as follows:
;
; Procedures (event handlers) are associated with events and objects in a 
; dictionary (a dynarray) in the following general form:
;
;         EventMan.Dictionary[ObjectTag + EventTag] = ProcName
;
; The idea is to EXECPROC ProcName based on the current object and event.
; 
; EventMan organizes programming in Paradox 4.0 into 4 phases:
;         1. Event collection
;         2. Event analysis
;         3. Event dispatch
;         4. Event handling
;
; Event collection can take place under either WAIT ... PROC or under 
; GETEVENT. EventMan.DoWait() and EventMan.DoGetEvent() are provided as 
; defaults for this.
;
; Event analysis involves several steps:
;         1. Get target window handle (placed in EventMan.TargetWindow)
;         2. Set object tag (placed in EventMan.ObjectTag)
;         3. Set event tag (placed in EventMan.EventTag)
;
; EventMan automatically places the target window handle in 
; EventMan.TargetWindow using EventMan.GetTargetWindow(..))
;
; EventMan provides a default procedure in EventMan.SetObjectTagProc
; (EventMan.SetWindowAsObjectTag()) for setting EventMan.ObjectTag. This
; procedure simply sets EventMan.ObjectTag to StrVal(EventMan.TargetWindow).
;
; Your application can substitute a more sophisticated procedure which might,
; for instance, set EventMan.ObjectTag to "INVOICE". 
; EventMan.SetIndexedObjectTag() is provided for this purpose. It looks
; up your custom ObjectTag from a list of window numbers. With this
; alternative, your application must maintain the cross referencing
; information in the dynarrays EventMan.WindowObjectTag[] and
; EventMan.ObjectTagWindow[]. The advantage to this is that by avoiding the
; use of window handles for ObjectTags, EventMan.Dictionary[] updates are not
; required each time a window is loaded or unloaded.
;
; EventMan sets EventMan.EventTag according to some rules that are easily
; understood by reading the code of the procedure EventMan.SetEventTag()
; below.
;
; Finally EventMan dispatches the event. EventMan provides a default
; dispatcher (EventMan.DispatchEventTag()) in EventMan.DispatchEventTagProc,
; although your application can substitute its own more sophisticated
; dispatcher (advanced applications only). The default dispatcher first
; looks for
;
;       EventMan.Dictionary[EventMan.ObjectTag + EventMan.EventTag]
;
; and then
;
;       EventMan.Dictionary[EvenMan.EventTag]
;
; if it didn't find the first form. Therefore default behavior can be
; provided by your application.
;
; Event handlers, the procedures that are EXECPROC'd by the event dispatcher,
; must be provided by your application, and must always return the values 0,
; 1, or 2, according to the rules provided for WAIT PROC commands.
;
; Special support is provided for IDLE event processing, and for MOUSE event
; processing, as well as populating and de-populating EventMan.Dictionary.
;
; Look for event packet details in EventMan.EventBag[].
; 
; For some ideas on how to use EventMan, see EDEMO.SC.
; 
;============================================================================
;                            EVENTMAN INTERFACE
;============================================================================
;
; Construction and destruction:
; -----------------------------
; EventMan.Constructor() ; execute at start of session
; EventMan.Destructor() ; execute at end of session
;
; The event dictionary and idle proc queue:
; -----------------------------------------
; EventMan.SetHandlersFrom(HandlerBag)
; Eventman.ClearHandlersWith(HandlerBag)
; EventMan.RegisterIdleServer(ObjectTag,IdleProcName)
; EventMan.DeRegisterIdleServer(ObjectTag)
;
; Event collectors: (use these in your code)
; ------------------------------------------
; (For Wait statements...)
; EventMan.CatchTrigger(EventMan.TriggerTag,EventMan.EventBag,
;       EventMan.TriggerCycle)
; (Shell (analog to wait) for EventMan.GetEventProc...)
; EventMan.DoGetEvent()
;
; Event analysis:
; ---------------
; EventMan.GetTargetWindow(TriggerTag,EventBag) ; returns 0 or valid window
; EventMan.SetWindowAsObjectTag() ; Default EventMan.SetObjectTagProc
; EventMan.SetIndexedObjectTag() ; More sophisticated object identification
; EventMan.SetEventTag() ; sets EventMan.EventTag from EventMan.EventBag[]
;
; Event dispatch:
; ---------------
; EventMan.DispatchEventBag() ; used by collectors
; EventMan.DispatchEventTag() ; default EventMan.DispatchEventTagProc; called
;                             ; by DispatchEventBag
; EventMan.DoIdleEvent()      ; execprocs next idle server
;
; Utility procs:
; ---------------------
; EventMan.DoWait() ; default wait statement, uses EventMan.CatchTrigger()
; EventMan.GetEvent() ; default EventMan.GetEventProc
; EventMan.EnableEvent() ; returns 0
; EventMan.DisableEvent() ; returns 1
; EventMan.CompleteMouseSequence() ; call after ExecEvent EventBag
; EventMan.ReturnFromMouseMode() ; call after mouse down when UP is available
;
; Properties:
; ------------
; EventMan.IsActive ; readonly, is EventMan active? (use isAssigned)
; EventMan.IsGetEvent ; readonly, is this a getevent event? (use isAssigned)
; EventMan.IsWait ; readonly, is this a wait event? (use isAssigned)
; EventMan.Dictionary[] ; EventMan.Dictionary[ObjectTag + EventTag] = ProcName
; EventMan.InterruptBag[] ; 
; EventMan.WindowObjectTag[] ;
; EventMan.ObjectTagWindow[] ;
; EventMan.EventBag[] ; as collected by wait or getevent
; EventMan.EventType ; TRIGGER or EventMan.EventBag["Type"]
; EventMan.MouseModeButton ; Button pressed with last DOWN, or "NONE"
; EventMan.TargetWindow ; set by collectors from EventMan.GetTargetWindow(..)
; EventMan.ObjectTag ; string value for EventMan.Dictionary set by 
;                    ; EventMan.SetWindowAsObjectTag() or EventMan.SetObjectTagProc
; EventMan.EventTag ; sometimes composite string value for 
;                   ; EventMan.Dictionary set by EventMan.SetEventTag()
; EventMan.ReturnCode ; must be set by DispatchEventTagProc
; EventMan.DefaultReturnCode ; return code when handler search fails
; EventMan.GetEventProc ; getevent specs for DoGetEvent
; EventMan.SetObjectTagProc ; sets EventMan.ObjectTag based on 
;                           ; Eventman.TargetWindow
; EventMan.DispatchEventTagProc ; dispatches event to EventMan.Dictionary
; 
;============================================================================
;                         EVENTMAN IMPLEMENTATION
;============================================================================

;============================================================================
;                         CONSTRUCTION AND DESTRUCTION
;============================================================================
Proc EventMan.Constructor()
;   If IsAssigned(EventMan.nInstances) Then
;      EventMan.nInstances = EventMan.nInstances + 1
;      Return
;   Endif
   ;-------------------------------------------------------------------------
   ; General properties
   ;-------------------------------------------------------------------------
   EventMan.IsActive = True
;   EventMan.nInstances = 1
   Dynarray EventMan.Dictionary[]
   Dynarray EventMan.EventBag[]
   Dynarray EventMan.WindowObjectTag[]
   Dynarray EventMan.ObjectTagWindow[]
   Dynarray EventMan.InterruptBag[]
   EventMan.MouseModeButton = "NONE"
   EventMan.EventType = ""
   EventMan.EventTag = ""
   EventMan.ObjectTag = ""
   EventMan.ReturnCode = 0
   EventMan.DefaultReturnCode = 0
   ;-------------------------------------------------------------------------
   ; Default action procs   
   ;-------------------------------------------------------------------------
   EventMan.GetEventProc = "EventMan.GetEvent"
   EventMan.SetObjectTagProc = "EventMan.SetWindowAsObjectTag"
   EventMan.DispatchEventTagProc = "EventMan.DispatchEventTag"
   ;-------------------------------------------------------------------------
   ; Private attributes
   ;-------------------------------------------------------------------------
   Dynarray EventMan_IdleServers[]
   Array EventMan_IdleServerProcs[1] ; stub
   EventMan_nIdleServers = 0
   EventMan_IdleServerPtr = 1

EndProc ; EventMan.Constructor

Proc EventMan.Destructor()
;   If EventMan.nInstances > 1 Then
;      EventMan.nInstances = EventMan.nInstances - 1
;      Return
;   Endif
   Release Vars

      EventMan.IsActive,
;      EventMan.nInstances,
      EventMan.IsWait,
      EventMan.IsGetEvent,
      EventMan.Dictionary,
      EventMan.EventBag,
      EventMan.WindowObjectTag,
      EventMan.ObjectTagWindow,
      EventMan.MouseModeButton,
      EventMan_IdleServers,
      EventMan_IdleServerProcs,
      EventMan_nIdleServers,
      EventMan_IdleServerPtr,
      EventMan.EventType,
      EventMan.EventTag,
      EventMan.ObjectTag,
      EventMan.TargetWindow,
      EventMan.ReturnCode,
      EventMan.DefaultReturnCode,
      EventMan.GetEventProc,
      EventMan.SetObjectTagProc,
      EventMan.DispatchEventTagProc

EndProc ; EventMan.Destructor

;============================================================================
;                         DICTIONARY UPDATE
;============================================================================

Proc EventMan.SetHandlersFrom(HandlerBag)
   Private
      EventTag,
      ObjectTag

   If IsAssigned(HandlerBag["ObjectTag"]) Then
      ObjectTag = Strval(HandlerBag["ObjectTag"])
      Release Vars HandlerBag["ObjectTag"] ; to avoid screening below
   Else
      ObjectTag = ""
   Endif

   Foreach EventTag in HandlerBag
      If (HandlerBag[EventTag] = "") Then
         Release Vars EventMan.Dictionary[ObjectTag + EventTag]
      Else
         EventMan.Dictionary[ObjectTag + EventTag] = HandlerBag[EventTag]
      Endif
   EndForeach

   If Not isBlank(ObjectTag) Then
      HandlerBag["ObjectTag"] = ObjectTag
   Endif
EndProc ; EventMan.SetHandlersFrom

Proc EventMan.ClearHandlersWith(HandlerBag)
   Private
      EventTag,
      ObjectTag

   If IsAssigned(HandlerBag["ObjectTag"]) Then
      ObjectTag = Strval(HandlerBag["ObjectTag"])
      Release Vars HandlerBag["ObjectTag"]
   Else
      ObjectTag = ""
   Endif

   Foreach EventTag in HandlerBag
      Release Vars EventMan.Dictionary[ObjectTag + EventTag]
   EndForeach

   If Not IsBlank(ObjectTag) Then
      HandlerBag["ObjectTag"] = ObjectTag
   Endif

EndProc ; EventMan.ClearHandlersWith

;============================================================================
;                         IDLE SERVER UPDATE
;============================================================================

Proc EventMan.RegisterIdleServer(ObjectTag,IdleProcName)
   Private
      Element,
      i
   If IsAssigned(EventMan_IdleServers[ObjectTag]) Then
      EventMan_IdleServers[ObjectTag] = IdleProcName
   Else
      EventMan_IdleServers[ObjectTag] = IdleProcName
      If EventMan_nIdleServers = 0 Then
         EventMan.InterruptBag["Idle"] = "EventMan.DoIdleEvent"
      Endif
      EventMan_nIdleServers = EventMan_nIdleServers + 1
      Array EventMan_IdleServerProcs[EventMan_nIdleServers]
      i = 0
      ForEach Element In EventMan_IdleServers
         i = i + 1
         EventMan_IdleServerProcs[i] = EventMan_IdleServers[Element]
      EndForEach 
      If EventMan_IdleServerPtr > EventMan_nIdleServers Then
         EventMan_IdleServerPtr = EventMan_nIdleServers
      Endif
   Endif
EndProc ; EventMan.RegisterIdleServer

Proc EventMan.DeRegisterIdleServer(ObjectTag)
   Private
      Element,
      i
   Release Vars
      EventMan_IdleServers[ObjectTag]
      
   EventMan_nIdleServers = EventMan_nIdleServers - 1
   If EventMan_nIdleServers > 0 Then
      Array EventMan_IdleServerProcs[EventMan_nIdleServers]
      i = 0
      ForEach Element In EventMan_IdleServers
         i = i + 1
         EventMan_IdleServerProcs[i] = EventMan_IdleServers[Element]
      EndForEach 
      If EventMan_IdleServerPtr > EventMan_nIdleServers Then
         EventMan_IdleServerPtr = EventMan_nIdleServers
      Endif
   Else
      Release Vars EventMan.InterruptBag["Idle"]
      EventMan_IdleServerPtr = 1
   Endif
EndProc ; EventMan.DeRegisterIdleServer

;============================================================================
;                         EVENT COLLECTORS
;============================================================================

;----------------------------------------------------------------------------
; The event handlers Return 0, 1 or 2 to determine whether to
; 0 = process the event with execevent
; 1 = deny the event but stay in the event loop
; 2 = deny the event and break out of the event loop
;----------------------------------------------------------------------------
Proc EventMan.CatchTrigger(EventMan.TriggerTag,EventMan.EventBag,EventMan.TriggerCycle)
   Private
      EventMan.IsGetEvent, ; mask out EventMan.IsGetEvent so that called procedures
                       ; can determine that it is a wait and not a getevent
                       ; that is active.
      EventMan.IsWait
   EventMan.IsWait = True
   EventMan.TargetWindow = 
      EventMan.GetTargetWindow(EventMan.TriggerTag,EventMan.EventBag)
   ExecProc EventMan.SetObjectTagProc
   If EventMan.TriggerTag = "EVENT" Then
      EventMan.DispatchEventBag()
   Else
      EventMan.EventType = "TRIGGER"
      EventMan.EventTag = EventMan.TriggerTag
      ExecProc EventMan.DispatchEventTagProc
   Endif
   Return EventMan.ReturnCode
EndProc ; EventMan.CatchTrigger

Proc EventMan.DoGetEvent()
   Private
      EventMan.IsGetEvent,
      EventMan.IsWait
   EventMan.IsGetEvent = True
   EventMan.ReturnCode = 0
   While EventMan.ReturnCode < 2
      ExecProc EventMan.GetEventProc
      EventMan.TargetWindow = 
         EventMan.GetTargetWindow("EVENT",EventMan.EventBag)
      ExecProc EventMan.SetObjectTagProc
      EventMan.DispatchEventBag()
      If EventMan.ReturnCode = 0 Then
         ExecEvent EventMan.EventBag
      EndIf
   EndWhile
EndProc ; EventMan.DoGetEvent

;============================================================================
;                         EVENT ANALYSIS
;============================================================================

;----------------------------------------------------------------------------
; Assumes ECHO NORMAL
;----------------------------------------------------------------------------
Proc EventMan.GetTargetWindow(TriggerTag,EventBag)
   Private
      TargetWindow,
      WindowBag,
      EventType,
      ZOrderGroupChange,
      IsEditor,
      Keycode,
      WindowList,
      i,
      Found
   ;---------------------------------------------------------------------
   ; If the trigger is not an event then analyze trigger...
   ;---------------------------------------------------------------------
   If TriggerTag <> "EVENT" Then ; must be a TRIGGER type event
      ;------------------------------------------------------------------
      ; If the trigger is not an ARRIVEWINDOW, then it relates to the 
      ; current image. 
      ;------------------------------------------------------------------
      If TriggerTag = "ARRIVEWINDOW" Then
         TargetWindow = GetWindow()
      Else
         Window Handle Form to TargetWindow
         If Not isWindow(TargetWindow) Then
            Window Handle Image ImageNo() To TargetWindow
         Endif
      Endif
   ;---------------------------------------------------------------------
   ; ...else this must be an event of type MOUSE, KEY, MESSAGE, or IDLE...
   ;---------------------------------------------------------------------
   Else
      If Not IsWindow(GetWindow()) Then ; There are no windows
         TargetWindow = GetWindow()
      Else
         EventType = EventBag["Type"]
         Switch
            Case EventType = "MOUSE":
               ;----------------------------------------------------------------
               ; If this is a mouse event, then if it is a DOWN or no-button
               ; event then it belongs to the window over which the mouse 
               ; event takes place, otherwise it is a modal event which 
               ; belongs to the current window.
               ;----------------------------------------------------------------
               If (EventBag["Action"] = "DOWN") or
                  ((EventBag["Buttons"] = "NONE") And
                  (EventBag["Action"] <> "UP")) Then
                  TargetWindow = WindowAt(EventBag["Row"],EventBag["Col"])
               Else
                  TargetWindow = GetWindow()
               Endif
               ;----------------------------------------------------------------
               ; Except for one special case when the mouse action is DOWN:
               ; If the current window is floating and the target window is not
               ; then expect to select the topmost desktop window as the target
               ; in the course of changing z-orders from floating to desktop,
               ; if the mouse event took place somewhere other than on the menu
               ; bar or the status bar.
               ;----------------------------------------------------------------
               If EventBag["Action"] = "DOWN" Then
                  Window GetAttributes GetWindow() To WindowBag
                  If WindowBag["Floating"] Then
                     If IsWindow(TargetWindow) Then
                        Window GetAttributes Targetwindow To WindowBag
                        ZOrderGroupChange = (Not WindowBag["Floating"])
                     Else ; see if the mouse hit the prompt/speedbar area...
                        SysInfo To SysInfoBag
                        ZOrderGroupChange = 
                           ((SysInfoBag["ScreenHeight"] - 1 <> EventBag["Row"]) And
                           (EventBag["Row"] <> 0))
                     Endif ; The target window exists
                     If ZOrderGroupChange Then
                        Window List To WindowList
                        Found = False
                        For i From 1 To Arraysize(WindowList)
                           Window GetAttributes WindowList[i] To WindowBag
                           If Not WindowBag["Floating"] Then
                              Found = True
                              Quitloop
                           Endif
                        EndFor
                        TargetWindow = IIf(Found,WindowList[i],0)
                     Endif
                  Endif ; The current window is a floating window
               Endif ; If this is a DOWN action
            Case EventType = "KEY":
               Keycode = EventBag["KeyCode"]
               Switch
                  Case KeyCode = Asc("WinResize") or
                     KeyCode = Asc("WinMax") or
                     KeyCode = Asc("WinNext") or
                     KeyCode = Asc("WinClose"):
                     TargetWindow = GetWindow()
                  Case (Keycode = Asc("ClearImage") or
                     Keycode = Asc("ClearAll") or
                     Keycode = Asc("EditKey") or
                     Keycode = Asc("CoEditKey")):
                     If nImages() > 0 Then
                        Window Handle Form to TargetWindow
                        If Not IsWindow(TargetWindow) Then
                           Window Handle Image ImageNo() To TargetWindow
                        Endif
                     Else
                        TargetWindow = 0
                     Endif
                  Otherwise:
                     ;----------------------------------------------------------
                     ; Check to see if the keystrokes belong to a current file or
                     ; memo editor...
                     ;----------------------------------------------------------
                     IsEditor = (IsFieldView() and Substr(FieldType(),1,1) = "M") 
                     or SysMode() = "File Editor" or SysMode() = "Script"
                     ;----------------------------------------------------------
                     ; If it does belong to an editor, then find the first non-
                     ; floating window on the desktop. That will be the editor.
                     ;----------------------------------------------------------
                     If IsEditor Then
                        Window List To WindowList
                        For i From 1 to ArraySize(WindowList)
                           Window GetAttributes WindowList[i] To WindowBag
                           If Not WindowBag["Floating"] Then
                              TargetWindow = WindowList[i]
                              QuitLoop
                           Endif
                        EndFor
                     Else ; Canvas or image is top window
                        ;-------------------------------------------------------
                        ; if there is an image then it will get the keystroke...
                        ;-------------------------------------------------------
                        If nImages() > 0 Then 
                           Window Handle Form To TargetWindow
                           If Not IsWindow(TargetWindow) Then
                              Window Handle Image ImageNo() To TargetWindow
                           Endif
                        Else
                           ;----------------------------------------------------
                           ; ...otherwise the desktop will get the keystroke
                           ;----------------------------------------------------
                           TargetWindow = 0
                        Endif
                     Endif
               EndSwitch
            Otherwise: ; MESSAGE or IDLE
               TargetWindow = GetWindow()
         EndSwitch
      Endif ; There is no window
   Endif ; Trigger
   Return TargetWindow
EndProc ; EventMan.GetTargetWindow

;----------------------------------------------------------------------------
; Sets ObjectTag to (the strval of)  the target window number, unless the
; target window number is 0, in which case sets ObjectTag o a blank string...
;----------------------------------------------------------------------------
Proc EventMan.SetWindowAsObjectTag()
   If EventMan.TargetWindow = 0 Then
      EventMan.ObjectTag = ""
   Else
      EventMan.ObjectTag = StrVal(EventMan.TargetWindow)
   Endif
EndProc ; EventMan.SetWindowAsObjectTag

Proc EventMan.SetIndexedObjectTag()
   If IsAssigned(EventMan.WindowObjectTag[EventMan.TargetWindow]) Then
      EventMan.ObjectTag = EventMan.WindowObjectTag[EventMan.TargetWindow]
   Else
      EventMan.ObjectTag = ""
   Endif
EndProc ; EventMan.SetIndexedObjectTag

Proc EventMan.SetEventTag()
   Eventman.EventType = EventMan.EventBag["Type"]
   Switch
      Case EventMan.EventType = "KEY":
         EventMan.EventTag = StrVal(EventMan.EventBag["Keycode"])
      Case EventMan.EventType = "MOUSE":
         Action = EventMan.EventBag["Action"]
         Switch
            Case Action = "DOWN":
               EventMan.EventTag = EventMan.EventBag["Buttons"] +
                  IIf(EventMan.EventBag["DoubleClick"],"DOUBLE","") +
                  Action
               EventMan.MouseModeButton = EventMan.EventBag["Buttons"]
            Case Action = "UP":
               EventMan.EventTag = 
                  IIf(EventMan.EventBag["DoubleClick"],"DOUBLE","") +
                  Action
            Case EventMan.EventBag["Buttons"] = "NONE":
               EventMan.EventTag = "NONE" + Action
               EventMan.MouseModeButton = "NONE"
            OtherWise:
               EventMan.EventTag = 
                   EventMan.EventBag["Buttons"] + Action
         EndSwitch
      Case EventMan.EventType = "MESSAGE":
         EventMan.EventTag = EventMan.EventBag["Message"]
         If EventMan.EventTag = "MENUSELECT" Then
            EventMan.EventTag = "MENUTAG:" + EventMan.EventBag["MenuTag"]
         Endif
         If EventMan.EventTag = "MENUKEY" Then
            EventMan.EventTag = "MENUKEY:" + EventMan.EventBag["Keycode"]
         Endif
      Case EventMan.EventType = "IDLE":
         EventMan.EventTag = "IDLE"
   EndSwitch
EndProc ; EventMan.SetEventTag

;============================================================================
;                         EVENT DISPATCH
;============================================================================

Proc EventMan.DispatchEventBag()
   Private
      OK,
      Action
   OK = False
   EventMan.SetEventTag()
   If IsAssigned(EventMan.InterruptBag[EventMan.EventTag]) Then
      ExecProc EventMan.InterruptBag[EventMan.EventTag]
   Else
      ExecProc EventMan.DispatchEventTagProc
   Endif
;   If EventMan.EventTag = "IDLE" Then
;      OK = EventMan.DoIdleEvent()
;   Endif
;   If Not OK Then
;      ExecProc EventMan.DispatchEventTagProc
;   Endif
EndProc ; EventMan.DispatchEventBag

;----------------------------------------------------------------------------
; Fires EventMan.Dictionary[EventMan.ObjectTag + EventMan.EventTag] if found,
; or EventMan.Dictionary[EventMan.EventTag] if found,
; or just returns (sets) EventMan.DefaultReturnCode
;----------------------------------------------------------------------------
Proc EventMan.DispatchEventTag()
   Switch
      Case EventMan.ObjectTag <> "" And 
         IsAssigned(EventMan.Dictionary[EventMan.ObjectTag +
                      EventMan.EventTag]):
         ExecProc EventMan.Dictionary[EventMan.ObjectTag +
                      EventMan.EventTag]
         EventMan.ReturnCode = Retval
      Case IsAssigned(EventMan.Dictionary[EventMan.EventTag]):
         ExecProc EventMan.Dictionary[EventMan.EventTag]
         EventMan.ReturnCode = Retval
      Otherwise:
         EventMan.ReturnCode = EventMan.DefaultReturnCode
   EndSwitch
EndProc ; EventMan.DispatchEventTag

Proc EventMan.DoIdleEvent()
   If EventMan_nIdleServers > 0 Then
      ExecProc EventMan_IdleServerProcs[EventMan_IdleServerPtr]
      If EventMan_IdleServerPtr = EventMan_nIdleServers Then
         EventMan_IdleServerPtr = 1
      Else
         EventMan_IdleServerPtr = EventMan_IdleServerPtr + 1
      Endif
   Endif
   EventMan.ReturnCode = 1
EndProc ; EventMan.DoIdleEvent

;============================================================================
;                       UTILITY PROCS
;============================================================================

;----------------------------------------------------------------------------
; Collects all events to EventMan.EventBag[]...
;----------------------------------------------------------------------------
Proc Eventman.DoWait()
   Echo Normal
   Wait WorkSpace
      Proc "EventMan.CatchTrigger" All
   EndWait
EndProc

;----------------------------------------------------------------------------
; Collects all events to EventMan.EventBag[]...
;----------------------------------------------------------------------------
Proc EventMan.GetEvent()
   GetEvent All To EventMan.EventBag
EndProc ; EventMan.GetEvent

;----------------------------------------------------------------------------
; Calling procedures should in most cases issue ExecEvent EventMan.EventBag
; to issue the mouse down event that initiated the sequence before calling
; EventMan.CompleteMouseSequence, to be modal. CompleteMouseSequence always
; returns 1 or 2. It does NOT refresh EventMan.TargetWindow nor 
; EventMan.ObjectTag because being modal it refers to the object or window
; that was current with the defining mouse down event.
;----------------------------------------------------------------------------
Proc EventMan.CompleteMouseSequence()
   EventMan.ReturnCode = 0
   While (EventMan.EventBag["Buttons"] <> "NONE")
      While True
         GetEvent to EventMan.EventBag
         If EventMan.EventBag["Type"] = "MOUSE" Then
            Quitloop
         Endif
      EndWhile
      EventMan.DispatchEventBag()
      If EventMan.ReturnCode = 0 Then
         ExecEvent EventMan.EventBag
      EndIf
   EndWhile
   If EventMan.ReturnCode = 0 Then
      EventMan.ReturnCode = 1
   Endif
   Return EventMan.ReturnCode
EndProc ; EventMan.CompleteMouseSequence

Proc EventMan.ReturnFromMouseMode()
   While (EventMan.EventBag["Buttons"] <> "NONE")
      While True
         GetEvent to EventMan.EventBag
         If EventMan.EventBag["Type"] = "MOUSE" Then
            Quitloop
         Endif
      EndWhile
   EndWhile
   EventMan.ReturnCode = 1
   Return EventMan.ReturnCode
EndProc ; EventMan.ReturnFromMouseMode

Proc EventMan.DisableEvent()
   Return 1
EndProc

Proc EventMan.EnableEvent()
   Return 0
EndProc

;============================================================================
;                         MAINLINE
;============================================================================
If Not IsAssigned(Librarian.HasControl) Then
   EventMan.Constructor()
   SetCanvas Default
   Echo Normal
   EventMan.DoGetEvent()
   EventMan.Destructor()
Endif
